/* qmdiritem.cpp
 *
 * $Id: qmdiritem.cpp,v 1.11 2000/12/09 12:12:36 amos Exp $
 *
 * QtVu is a user-friendly image viewer and browser for X11.
 * See http://www.qtvu.org.
 * Copyright(C) 2000 Apollo Team.  See CREDITS file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The GNU General Public License is also available online at:
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#include "qmdiritem.h"
#include "qmsongitem.h"

#include <iostream.h>

#include <qapplication.h>
#include <qdir.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qlistview.h>
#include <qpixmap.h>
#include <qstring.h>

#include "icons/folder.xpm"
#include <qglobal.h>

/*! \file qmdiritem.cpp
  Implementation of QmDirItem class which represents a directory in the directory tree browser.
*/

/*!
  \class QmDirItem qmdiritem.h
  \brief Represents a directory in the directory tree browser.

  Instances of this class represents directories in the directory
  (tree) browser.  A directory is loaded (listed) when it is opened.
*/

/* Based on the dirview example by Troll Tech. */

/*!
  Creates a new directory item.  \a parent is the parent directory item,
  and \a filename is the path to be read.  The \a readable and \a
  symlink arguments will affect how the item is going to look.
*/

QmDirItem::QmDirItem(
    QmDirItem *parent,
    const char *path,
    int count)
    : QmPlayListItem(parent, count),
    m_Dir(path)
{
	setPixmap( 0, QPixmap( folder_xpm ) );
	setOpen(true);
}


/*!
  Creates a new directory item.  The root will be listed.
 */

QmDirItem::QmDirItem(
    QListView *parent,
	const char *path,
    int count)
    : QmPlayListItem(parent, count),
    m_Dir(path)
{
	setPixmap( 0, QPixmap( folder_xpm ) );
	setOpen(true);
}


/*!
  Reads the sub directories of this directory.  If \a o is true,
  the tree will be exanded, otherwise collapsed.
 */

void
QmDirItem::setOpen(
    bool /*o*/)
{
    QListViewItem::setOpen(true);
	
    if( childCount() == 0 )
	{
//		qDebug( "QmDirItem::setOpen: %s", fullName().ascii() );
		
        qApp->processEvents();
		QString s(fullName());
        const QFileInfoList *filelist;
		
		QDir dir(s);
		dir.setFilter(QDir::Readable);
		dir.setSorting(QDir::DirsFirst);
		filelist = dir.entryInfoList( QDir::Dirs );

		if(filelist)
		{
			QFileInfoListIterator it(*filelist);
			QFileInfo *fi;
	    
			while((fi = it.current()))
			{
				++it;

//				cout << fi->fileName() << " - " << fi->dirPath(true) << endl;

				if(fi->fileName() != "." && fi->fileName() != ".." && fi->isReadable())
				{
					if(fi->isDir())
						(void) new QmDirItem( this, fi->absFilePath() );
                    else
                        qWarning( "\"%s\" is not a directory which was expected by file info",
                                  fi->absFilePath().ascii() );
                    qApp->processEvents();
				}
			}
		}

        dir = QDir(s);
		dir.setFilter(QDir::Readable);
		dir.setNameFilter("*.mp3");
		dir.setSorting(QDir::DirsFirst);
		filelist = dir.entryInfoList( QDir::Files );

		if(filelist)
		{
			QFileInfoListIterator it(*filelist);
			QFileInfo *fi;
	    
			while((fi = it.current()))
			{
				++it;

				if(fi->fileName() != "." && fi->fileName() != ".." && fi->isReadable())
				{
					if(!fi->isDir())
						(void) new QmSongItem( this, fi->absFilePath() );
                    else
                        qWarning( "\"%s\" is not a file which was expected by file info",
                                  fi->absFilePath().ascii() );
				}
			}
		}
    }
}


/*!
  If a directory does not contain any directories, it will not be expandable.
*/

void
QmDirItem::setup()
{
//      QDir d(fullName());
//      d.setFilter(QDir::Dirs);

//      // If there are no directories, 2 will be returned (for `.' and `..').
    
//      if(d.count() == 2)
//  	setExpandable(false);
//      else
	setExpandable(true);
	
    QListViewItem::setup();
}


/*!
  Returns the full path of the directory, including trailing slash.
 */

QString
QmDirItem::fullName()
{
	return m_Dir.absPath();
}


/*!

 */

QString
QmDirItem::text(
    int column) const
{
    if ( column == 0 )
		return m_Dir.dirName();
	else
		return "";
}


/*!
  \return The save string of the item.  This string can be saved to
  disk, and loaded at a later time to create an identical instance
  of this class.
*/

QString
QmDirItem::save() const
{
	return m_Dir.absPath();
}


/*!
  This function is used to determine whether an item is a song
  or a folder.

  \return False for items of this class.
*/

bool
QmDirItem::isSong() const
{
	return false;
}

QString
QmDirItem::key( int col, bool  ) const
{
    if ( depth() == 0 )
    {
        return sortCounterString();
    }
    else
    {
        if ( col == 0 )
            return QString( "a" ) + text( 0 );
        else
            return text( col );
    }
}
