/* qmmenubutton.cpp
 *
 * $Id: qmmenubutton.cpp,v 1.4 2000/12/09 12:12:36 amos Exp $
 *
 * <*** DESCRIPTION OF PROGRAM ***>
 * Copyright(C) 2000 Apollo Team.  See CREDITS file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The GNU General Public License is also available online at:
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#include "qmmenubutton.h"

#include <iostream>

#include <qiconset.h>
#include <qpainter.h>
#include <qevent.h>

/*! \file qmmenubutton.cpp
  Implementation of QmMenuButton class which displays a toolbutton with a popup menu.
*/

/*!
  \class QmMenuButton qmmenubutton.h
  \brief For creating buttons in toolbars that open a popup menu.

  This class acts as a QToolButton, except that it adds a small down
  arrow of the right of the text/icon.  Clicking the arrow will emit
  menuClicked() signal, while clicking on the icon will emit
  iconClicked().
*/


/*!
  Constructs a new menu button.
*/

QmMenuButton::QmMenuButton(
	const QIconSet & s,
	const QString & textLabel,
	QWidget * parent,
	const char * name)
	: QToolButton(parent, name),
	  m_MouseOver(false)
{
	setOnIconSet( s );
	setTextLabel( textLabel );
	
	m_PixmapWidth = s.pixmap(QIconSet::Small, QIconSet::Normal).width();
	setFixedWidth( m_PixmapWidth + 19 ); // space + arrowwidth + space + border
}


/*!

*/

QmMenuButton::~QmMenuButton()
{
}


/*!
  Reimplemented to draw down arrow and separator between icon and arrow.
 */

void
QmMenuButton::drawButton(
	QPainter *p)
{	
    style().drawToolButton( this, p );
	
	QPixmap pixmap = onIconSet().pixmap(QIconSet::Small, QIconSet::Normal);
	p->drawPixmap(3, 3, pixmap);
	style().drawArrow(p, DownArrow, true,
					  pixmap.width() + 10,
					  height() / 2 - 2, 5, 5,
					  colorGroup(), isEnabled());
	
	if(m_MouseOver)
	{
		if(isDown())
		{
			p->setPen(colorGroup().light());
			p->drawLine(pixmap.width() + 5, 1, pixmap.width() + 5, height() - 2);
			p->setPen(colorGroup().dark());
			p->drawLine(pixmap.width() + 6, 1, pixmap.width() + 6, height() - 2);
		}
		else
		{
			p->setPen(colorGroup().dark());
			p->drawLine(pixmap.width() + 5, 1, pixmap.width() + 5, height() - 2);
			p->setPen(colorGroup().light());
			p->drawLine(pixmap.width() + 6, 1, pixmap.width() + 6, height() - 2);
		}
	}
}


/*!
  Keeps state so that the button is drawn correctly.
 */

void
QmMenuButton::enterEvent(
	QEvent *e)
{
	m_MouseOver = true;
	QToolButton::enterEvent(e);
}


/*!
  Keeps state so that the button is drawn correctly.
 */

void
QmMenuButton::leaveEvent(
	QEvent *e)
{
	m_MouseOver = false;
	QToolButton::leaveEvent(e);
}


/*!
  Emits menuClicked() is the arrow region was clicked.  If the
  icon region was clicked, iconClicked() will be emitted.
 */

void
QmMenuButton::mouseReleaseEvent(
	QMouseEvent *e)
{
	if(m_MouseOver)
	{
		if(e->x() > m_PixmapWidth + 6)
			emit menuClicked();
		else
			emit iconClicked();
	}

	QToolButton::mouseReleaseEvent(e);
}

