/* qmmixer.cpp
 *
 * $Id: qmmixer.cpp,v 1.3 2000/12/09 12:12:36 amos Exp $
 *
 * Apollo sound player: http://www.apolloplayer.org
 * Copyright(C) 2000 Apollo Team.  See CREDITS file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The GNU General Public License is also available online at:
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#include "qmmixer.h"

#include <fcntl.h>
#include <iostream.h>
#include <linux/soundcard.h>
#include <math.h>
#include <sys/ioctl.h>

/*! \file qmmixer.cpp
  Implementation of QmMixer class which provides functionality to control the mixer.
*/

/*!
  \class QmMixer qmmixer.h
  \brief Provides functionality to control the mixer.

  This class is used to adjust the volume of the mixer.
*/


/*!
  Constructs a new mixer control.  Set \a logarithmic to
  true to enable logarithmic volume control.  If set to
  false (default), linear volume control will be used.

  The choice is given because some sound cards do not
  seem to support the logarithmic control.  Logarithmic
  control may give a more pleasant result for controlling
  the volume, though, as a linear control may not \e
  sound linear, i.e. 50% of the volume in linear mode
  may seem like 30% of the volume.
*/

QmMixer::QmMixer(
	bool logarithmic)
	: m_UseLog(logarithmic)
{
	m_MixerFd = open("/dev/mixer", O_RDWR);

	if(m_MixerFd == -1)
	{
		cout << "Can't open mixer" << endl;
		// do exception or something
	}
}


/*!

*/

QmMixer::~QmMixer()
{
	// For some reason, closing the mixer devices gives core dump.
	// Sample mixer programs doesn't seem to close the mixer fd.
//	close(m_MixerFd);
}


/*!
  Reads the current volume of PCM from the mixer device.
 */

int
QmMixer::volume() const
{
	int vol;
	int res = ioctl(m_MixerFd, MIXER_READ(SOUND_MIXER_PCM), &vol);

	if(res == -1)
	{
		// undefined mixer channel was accessed
		cout << "Error reading volume: undefined channel." << endl;
	}

	// Least significant byte is left channel volume.  Next byte is right channel volume.
	// The upper 16 bits are undefined.  For mono devices, only the left channel is used.

//	int left = vol & 0x7f;
//	int right = (vol >> 8) & 0x7f;
//	cout << " left channel is: " << left << endl
//		 << "right channel is: " << right << endl;

	// Since we don't offer individual left/right volume control, we'll simply use
	// the left channel (as this is also used for mono devices) as the volume, and
	// set the right channel to the same value.
	int volume = vol & 0x7f;

	return volume;
}


/*!
  Sets the PCM volume of the mixer device to \a v.
 */

void
QmMixer::setVolume(
	int v)
{
	if(v == 0)
		v = 1;
	
	// Use same volume for left and right channels

	if(m_UseLog)
		v = (int)floor((log10((double)v) * 50.0 + 0.5));
	
	int vol = ((v << 8) & 0x7f00) | (v & 0x007f);
	int res = ioctl(m_MixerFd, MIXER_WRITE(SOUND_MIXER_PCM), &vol);

	if(res == -1)
	{
		cout << "Error setting volume" << endl;
	}
}
