/* qmmpg123player.cpp
 *
 * $Id: qmmpg123player.cpp,v 1.24 2000/12/18 17:55:21 amos Exp $
 *
 * Apollo sound player: http://www.apolloplayer.org
 * Copyright(C) 2000 Apollo Team.  See CREDITS file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The GNU General Public License is also available online at:
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <errno.h>
#include <iostream.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>

#include <qapplication.h>
#include <qfileinfo.h>
#include <qmessagebox.h>
#include <qsocketnotifier.h>
#include <qtextstream.h>
#include <qtimer.h>

#include "qmmpg123player.h"
#include "qmconfig.h"

bool 	QmMpg123Player::s_OurTermSignal;

/*! \file qmmpg123player.cpp
  Implementation of QmMpg123Player class which provides functionality for
  communicating with mpg123.
*/

/*!
  \class QmMpg123Player qmmpg123player.h
  \brief Provides functionality for communicating with mpg123.

  This class is used to communicate with <a href="http://www.mpg123.com">mpg123</a>
  to play MP3 files.  The communication is done through pipes.  Apollo sends a set
  of commands to mpg123, and mpg123 provides Apollo with playing info.

  Some of the code in this class was taken from Andreas Neuhaus' IRMP3.
  
  \warning The current mpg123 version does not handle all MP3 files very well.  With
  some MP3 files, mpg123 will provide faulty playing info but play the song correctly.
  The user interface will try to detect this, but may still act a little weird.

  \todo Add possibility to use buffer in mpg123 (-b).  The default is no buffering,
  which will cause the player to jump if the load is high.
*/


/*!

*/

QmMpg123Player::QmMpg123Player() : QmPlayer(0, 0),
                                   m_ID3Format( "%1 - %2" ),
                                   m_Playing( false ),
                                   m_Paused( false ),
                                   m_NewSong( false ),
                                   m_GotNewSong( false ),
                                   m_BufferLeftovers( 0 ),
                                   m_SendFd( 0 ),
                                   m_RecvFd( 0 ),
                                   m_Pid( 0 ),
                                   m_FrameCount( 0 ),
                                   m_Seconds( 0 )
{
	s_OurTermSignal = false;
    m_pTimer = new QTimer( this );
    connect( m_pTimer, SIGNAL( timeout() ),
             this, SLOT( releaseTimingInformation() ) );
}


/*!
  Sends a termination signal to the mpg123 process.
*/

QmMpg123Player::~QmMpg123Player()
{
//	strcpy(m_WriteBuffer, "QUIT");
//	sendCommand( m_WriteBuffer );	
//	kill(m_Pid, SIGQUIT);
	s_OurTermSignal = true;
	kill(m_Pid, SIGTERM);
}


/*!
  Starts mpg123.  If a mpg123 process is already running, it will
  be terminated and another one restarted.
*/

int
QmMpg123Player::init()
{
	// check if player is already running
	if (m_Pid)
	{
		s_OurTermSignal = true;
		kill(m_Pid, SIGKILL);
		delete m_pReadNotifier;
		close(m_SendFd);
		close(m_RecvFd);
	}

	// make socketpair to communicate with player
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd_send) < 0)
		return -1;
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd_recv) < 0)
		return -1;

	// handle signals
	signal(SIGPIPE, SIG_IGN);
	signal(SIGCHLD, QmMpg123Player::childHandler);
	signal(SIGUSR1, QmMpg123Player::failHandler);

	int apollopid = getpid();
	m_Pid = fork();
	
	if (m_Pid == 0)
	{
		// pipe in/output through socket to parent
		dup2(fd_send[0], STDIN_FILENO);
		close(fd_send[0]);
		close(fd_send[1]);
		dup2(fd_recv[0], STDOUT_FILENO);
		dup2(fd_recv[0], STDERR_FILENO);
		close(fd_recv[0]);
		close(fd_recv[1]);

		// Spawn player.  -R for starting mpg123 in remote mode (i.e. we're using
		// mpg123 through commands via pipe).  The last argument, '-', is necessary
		// otherwise mpg123 will quit immediately (dash is an alternative to an
		// URL or file name).
        QString mpgpath = QmConfig::instance()->mpg123Path();
		execlp(mpgpath.ascii(), mpgpath.ascii(), "-R", "-", 0);

		// never reached if exec was ok
		kill(apollopid, SIGUSR1);
		_exit(-1);
	}

	// parent continues here
	close(fd_send[0]);
	m_SendFd = fd_send[1];
	close(fd_recv[0]);
	m_RecvFd = fd_recv[1];

	m_pReadNotifier = new QSocketNotifier(m_RecvFd, QSocketNotifier::Read);
	QObject::connect(m_pReadNotifier, SIGNAL(activated(int)), this, SLOT(dataReceived()));
	
	FD_SET(m_RecvFd, &mpg123_fdset);

	s_OurTermSignal = false;	
	
	return 0;
}


/*!
  Plays the song \a Filename.
*/

void
QmMpg123Player::play(
    const QString &Filename)
{
    QFileInfo fi( m_Filename );
    m_FrameCount = -1;
    m_Seconds = -1;
    m_NewSong  		= true;
    m_GotNewSong    = false;
    m_Filename 		= Filename;
    m_FilePath 		= fi.absFilePath();
    m_FileExtension = fi.extension();
    sprintf(m_WriteBuffer, "LOAD %s", (const char*)Filename);
	sendCommand( m_WriteBuffer );
    startTimer();
}



void
QmMpg123Player::play()
{
    m_FrameCount = -1;
    m_Seconds = -1;
    m_NewSong = true;
    m_GotNewSong = false;
	restart();
    startTimer();
}


void
QmMpg123Player::startTimer()
{
    m_pTimer->start( 500, false );
}

/*!
  Stops the player.
*/

void
QmMpg123Player::stop()
{
    m_FrameCount = -1;
    m_Seconds = -1;
	strcpy(m_WriteBuffer, "STOP");
	sendCommand( m_WriteBuffer );
	m_Playing = false;
    m_pTimer->stop();
}


/*!
  Pauses the player.
*/

void
QmMpg123Player::pause()
{
	strcpy(m_WriteBuffer, "PAUSE");
	sendCommand( m_WriteBuffer );
}


/*!
  Restarts the song.
 */

void
QmMpg123Player::restart()
{
	strcpy(m_WriteBuffer, "JUMP 0");
	sendCommand( m_WriteBuffer );
}


/*!
  Reads the response from mpg123.
  \todo Later versions should use the un-terminated lines for the next decode run instead of ignoring them.
*/

void
QmMpg123Player::dataReceived()
{
//	memset(m_ReadBuffer, 0, READ_BUF_SIZE);
	int bytes_read = read(m_RecvFd,
                          m_ReadBuffer + m_BufferLeftovers,
                          READ_BUF_SIZE - m_BufferLeftovers );
	
	while(bytes_read > 0) // bytes_read == READ_BUF_SIZE
	{
		m_ReadBuffer[bytes_read] = '\0';
//          cout << "$$ " << m_ReadBuffer << " $$";

        char *buffer_pointers[50];
        char **buffer_ptr = buffer_pointers;

        // Split lines into separate strings and store their pointers
        char *tmp = m_ReadBuffer;
        char *last_tmp = tmp;
        while ( *tmp != 0 )
        {
            if ( *tmp == '\n' )
            {
                *tmp = 0;
                *buffer_ptr = last_tmp;
                ++buffer_ptr;
                last_tmp = tmp + 1;
            }
            ++tmp;
        }

        *buffer_ptr = 0; // This ends the pointer list

        buffer_ptr = buffer_pointers; // Start from the top

        char *read_buffer = m_ReadBuffer;
        while ( *buffer_ptr != 0 )
		{
            read_buffer = *buffer_ptr;
			if( read_buffer[0] == '@')
			{
				parseCodes( read_buffer );
			}
			else
			{
				cout << "dataReceived(): Received " << bytes_read << " bytes:" << endl
					 << read_buffer << endl;
				break;
			}

            // Jump to next string, if any
            ++buffer_ptr;
		}

        if ( tmp > last_tmp ) // There's some text without an endline
        {
            char *last = last_tmp;
            int leftovers = tmp - last_tmp;
            m_BufferLeftovers = leftovers;
            if ( leftovers > 0 ) // If non empty copy it to the beginning of the buffer
            {
                char *dst = m_ReadBuffer;
                for ( ; leftovers > 0; --leftovers )
                {
                    *dst++ = *last_tmp++;
                }
            }
            cout << "#" << last << "#\n";
        }
        else
            m_BufferLeftovers = 0;
			
		if(bytes_read != READ_BUF_SIZE)
			break;
		else
		{
			bytes_read = read(m_RecvFd,
                              m_ReadBuffer + m_BufferLeftovers,
                              READ_BUF_SIZE - m_BufferLeftovers );
			cout << "next set contains " << bytes_read << endl;
		}
	}
}


/*!
  Parses the response \a read_buffer from mpg123.
*/

void
QmMpg123Player::parseCodes( char *read_buffer )
{
    char code = read_buffer[1];
    switch(code)
    {
        case 'F':
            parseFrameInfo( read_buffer );
            break;

        case 'S':
            parseStatus( read_buffer );
            break;

        case 'I':
            parseInformation( read_buffer );
            break;

        case 'P':
            parsePlayStatus( read_buffer );
            break;

        case 'E':
            parseError( read_buffer );
            break;

        case 'R':
//              cout << "Version: " << read_buffer << endl;
            /*
              @R MPG123
              Startup version message
            */
            break;

        default:
            cout << "Unknown code: " << read_buffer << endl;
            break;
    }
}

/*!
  Status message during playing (frame info)
  @F <a> <b> <c> <d>
  a = framecount (int)
  b = frames left this song (int)
  c = seconds (float)
  d = seconds left (float)
  
  Examples:
	  
  @F 1249 9960 32.63 260.18
  @F 1250 9959 32.65 260.15
  @F 1251 9958 32.68 260.13
  @F 1252 9957 32.71 260.10
*/

void
QmMpg123Player::parseFrameInfo( char *read_buffer )
{
	int framecount, framesleft;
	float seconds, secondsleft;
	sscanf(&read_buffer[3], "%d %d %f %f", &framecount,
		   &framesleft, &seconds, &secondsleft);

    bool release = false;

	if(m_GotNewSong)
	{
		m_Playing = true;
        m_NewSong = false;
        m_GotNewSong = false;
		emit frameCount( framesleft + framecount );
        emit secondsCount((int)( seconds + secondsleft ));
        release = true;
	}

	if(framesleft == 0)
	{
		m_Playing = false;
	}

    m_FrameCount = framecount;
    m_Seconds = (int)seconds;

    if ( release ) // Is true if this is the first time, if so we need to send the information immediately
        releaseTimingInformation();
}


void QmMpg123Player::releaseTimingInformation()
{
    if ( m_Playing ) // Only emit while we are playing
    {
        emit framesLeft(m_FrameCount);
        emit secondsPlaying(m_Seconds);
    }
}


/*!
  @S <a> <b> <c> <d> <e> <f> <g> <h> <i> <j> <k> <l>
  Status message after loading a song (stream info)
  a = mpeg type (string)
  b = layer (int)
  c = sampling frequency (int)
  d = mode (string)
  e = mode extension (int)
  f = framesize (int)
  g = stereo (int)
  h = copyright (int)
  i = error protection (int)
  j = emphasis (int)
  k = bitrate (int)
  l = extension (int)
*/

void
QmMpg123Player::parseStatus( char * /*read_buffer*/ )
{
//  	cout << "Status: " << read_buffer << endl;
}


/*!
  @I ID3:<a><b><c><d><e><f>
  Status message after loading a song (ID3 song info)
  a = title (exactly 30 chars)
  b = artist (exactly 30 chars)
  c = album (exactly 30 chars)
  d = year (exactly 4 chars)
  e = comment (exactly 30 chars)
  f = genre (string)

  @I <a>
  Status message after loading a song (no ID3 song info)
  a = filename without path and extension
*/

void
QmMpg123Player::parseInformation( char *read_buffer )
{
    m_NewSong = true;
    m_GotNewSong = true;
    if ( read_buffer[3] == 'I' &&
         read_buffer[4] == 'D' &&
         read_buffer[5] == '3' &&
         read_buffer[6] == ':' )
    {
        char *info = &read_buffer[7];
        char title[31], artist[31], album[31], year[5], comment[31];
        char *genre;
        title[30] = artist[30] = album[30] = year[4] = comment[30] = 0;

        strncpy( title, info, 30 );
        info += 30;
        strncpy( artist, info, 30 );
        info += 30;
        strncpy( album, info, 30 );
        info += 30;
        strncpy( year, info, 4 );
        info += 4;
        strncpy( comment, info, 30 );
        info += 30;
        genre = info;

        m_Title = QString( title ).stripWhiteSpace();
        m_Artist = QString( artist ).stripWhiteSpace();
        m_Album = QString( album ).stripWhiteSpace();
        m_Year = QString( year ).stripWhiteSpace();
        m_Comment = QString( comment ).stripWhiteSpace();
        m_Genre = QString( genre ).stripWhiteSpace();

        QString song_title = resolveTags();

        emit songTitle( song_title );
    }
    else
    {
        char *info = &read_buffer[3];

        emit songTitle( info );
    }
}

QString
QmMpg123Player::resolveTags()
{
    QString song_title;
    int old_pos = 0;
    while ( old_pos < (int)m_ID3Format.length() && old_pos >= 0 )
    {
        int pos = m_ID3Format.find( '%', old_pos );
        if ( pos != -1 )
        {
            if ( pos != (int)m_ID3Format.length() -1 )
            {
                song_title += m_ID3Format.mid( old_pos, pos - old_pos );
                QChar c = m_ID3Format[pos+1];
                if ( c == '1' ) // Artist
                    song_title += m_Artist;
                else if ( c == '2' ) // Title
                    song_title += m_Title;
                else if ( c == '3' ) // Album
                    song_title += m_Album;
                else if ( c == '4' ) // Year
                    song_title += m_Year;
                else if ( c == '5' ) // Comment
                    song_title += m_Comment;
                else if ( c == '6' ) // Genre
                    song_title += m_Genre;
                else if ( c == '7' ) // File name
                    song_title += m_Filename;
                else if ( c == '8' ) // File path
                    song_title += m_FilePath;
                else if ( c == '9' ) // File extension
                    song_title += m_FileExtension;
                else
                    song_title += QString( "%" ) + c;
                old_pos = pos + 2;
            }
            else
            {
                song_title += m_ID3Format.mid( old_pos );
                old_pos = -1;
            }
        }
        else
        {
            song_title += m_ID3Format.mid( old_pos );
            old_pos = pos;
        }
    }
    return song_title;
}


/*!
  @P <a>
  Playing status
  a = 0: playing stopped
  a = 1: playing paused
  a = 2: playing unpaused
*/

void
QmMpg123Player::parsePlayStatus( char *read_buffer )
{
	int status;
	sscanf(&read_buffer[3], "%d", &status);

	switch(status)
	{
	case 0:
//  		cout << "stopped" << endl;
		emit playingStopped();
		break;

	case 1:
//  		cout << "paused" << endl;
		break;

	case 2:
//  		cout << "unpaused" << endl;
		break;

	default:
		cout << "Unknown play status: " << read_buffer << endl;
	}
	
//	emit statusChanged();
}


/*!
  @E <a>
  An error occured
  Errors may be also reported by mpg123 through
  stderr (without @E)
  a = error message (string)
*/

void
QmMpg123Player::parseError( char *read_buffer )
{
	cout << "Error: " << read_buffer << endl;
}


/*!
  \return True if playing, false otherwise.
 */

bool
QmMpg123Player::isPlaying() const
{
	return m_Playing;
}


/*!
  \return True if paused, false otherwise.
 */

bool
QmMpg123Player::isPaused() const
{
	return m_Paused;
}


/*!
  Sends the string \a write_buffer to the mpg123 process.  If
  the command was not sent successfully, a dialogbox will be
  shown.  This kind of failure will typically happen when the
  mpg123 process crashed, and the pipe is broken.

  \return True if successful, false otherwise.
*/

bool
QmMpg123Player::sendCommand(
	const char *write_buffer )
{
    int rc = write(m_SendFd, write_buffer, strlen(write_buffer));

	if(rc > -1)
		return true;
	else
	{
		rc = QMessageBox::warning(0,
								  QObject::tr("Apollo Error"),
								  QObject::tr("Apollo failed to send the command to<br>"
											  "the MP3 player.  This typically means<br>"
											  "that the MP3 player process is no longer<br>"
											  "working correctly and needs to be restarted.<br>"
											  "It might just have been a temporary problem<br>"
											  "and you can try again."),
								  QObject::tr("&Restart player"),
								  QObject::tr("&Do nothing"));

		if(rc == 0)
		{
			init();
			return sendCommand(write_buffer);
		}
	}

	return false;
}



/*!
  Signal handler for mpg123 process.
*/

/*
  According to the C++ FAQ [1], it is not a good idea to pass
  a pointer to a member function to a C function such as signal()
  for function callbacks.  Instead, we will have to rely on a
  static member function (which acts just as a regular C function).

  [1] See 30.2 at http://www.cs.uu.nl/wais/html/na-dir/C++-faq/part9.html
*/

void
QmMpg123Player::childHandler(
	int /*sig*/)
{
	// Catch signal of exiting player.

//	cout << "Signal from mpg123" << endl;

	if( ! s_OurTermSignal)
	{
		QMessageBox::critical(
			0,
			QObject::tr("Apollo Error"),
			QObject::tr("Apollo has detected that the MP3 player<br>"
						"process failed.  Apollo cannot function<br>"
						"without it.  You can try to restart the<br>"
						"player from the Options menu.<br>"
						"If the particular song you were trying to<br>"
						"play keeps continuing to fail the player<br>"
						"you can mark it as 'bad' from the Playlist menu."));
	}

	/*
    int status;
	
	if(waitpid(-1, &status, WNOHANG) == m_Pid)
	{
		cout << "mpg123 process exited with code " << WEXITSTATUS(status) << endl;
		m_Pid = 0;
	}
	*/
}


/*!
  Signal handler for mpg123 process.
*/

void
QmMpg123Player::failHandler(
	int /*sig*/)
{
	QMessageBox::critical(
		0,
		QObject::tr("Apollo Error"),
		QObject::tr("Apollo has detected that the MP3 player<br>"
					"is not running.  Apollo cannot function<br>"
					"without it.  You can try to restart the<br>"
					"player from the Options menu."));
}
