/* qmplaylist.cpp
 *
 * $Id: qmplaylist.cpp,v 1.16 2000/12/18 17:57:12 amos Exp $
 *
 * Apollo sound player: http://www.apolloplayer.org
 * Copyright(C) 2000 Apollo Team.  See CREDITS file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The GNU General Public License is also available online at:
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <iostream.h>

#include <qfile.h>
#include <qfileinfo.h>
#include <qtextstream.h>

#include "qmplaylist.h"
#include "qmsongitem.h"
#include "qmdiritem.h"

int QmPlayList::s_PlayListCounter = 0;

/*! \file qmplaylist.cpp
  Implementation of QmPlayList class which represents the playlist.
*/

/*!
  \class QmPlayList qmplaylist.h
  \brief Represents the playlist.

  This widget is the playlist in Apollo.  It provides functionality
  for loading and saving the playlist to file and traversing the
  playlist.
*/


/*!
  Constructs a playlist widget.
*/

QmPlayList::QmPlayList(
    QWidget *parent,
    const char *name)
    : QListView(parent, name),
	  m_pPlayingItem(0)
{
	// It is necessary to keep a 'play item' variable (see above) so we can
	// remove its playing status when another item is selected.
	// currentItem() cannot be used as it will already be updated with
	// the new song.
	
    setShowSortIndicator( true );

	// We need to maintain our playing song variable, so listen
	// to selection changes.
	
	connect(this, SIGNAL(doubleClicked(QListViewItem*)),
			this, SLOT(setPlaying(QListViewItem*)));
	connect(this, SIGNAL(returnPressed(QListViewItem*)),
			this, SLOT(setPlaying(QListViewItem*)));
}


/*!
  Destroys the playlist.  The playlist is \e not saved as we don't
  have a filename.
*/

QmPlayList::~QmPlayList()
{
}


/*!
  Saves the playlist to \a filename.  This function will silently
  skip saving if the playlist file cannot be created for some reason.

  \return True if the playlist was properly saved, false otherwise.

  \sa load(const QString &)
*/

bool
QmPlayList::save(
    const QString &filename)
{
    QFile f(filename);

    if(f.open(IO_WriteOnly))
    {
		QTextStream t(&f);
		t << filename << endl;

		QmPlayListItem *item = dynamic_cast<QmPlayListItem*>(firstChild());

		while(item != 0)
		{
			if(item->depth() == 0)
				t << item->save() << endl;
			item = dynamic_cast<QmPlayListItem*>(item->nextSibling());
		}
    }
    else
        return false;
    return true;
}


/*!
  Loads the playlist \a filename into the list view.   This function will silently
  skip loading if the playlist file cannot be loaded for some reason.

  \return True if the playlist was properly loaded, false otherwise.

  \sa save(const QString &)
*/

bool
QmPlayList::load(
	const QString &filename)
{
    QFile f(filename);
    QFileInfo f_inf(filename);
    QString abs_path = f_inf.dirPath( true );

    if(f.open(IO_ReadOnly))
    {
		QTextStream t(&f);
		QString s;

		s = t.readLine();

		while( ! t.atEnd())
		{
			s = t.readLine();
			QFileInfo fi(s);
            if ( fi.isRelative() ) // If it is relative we need to prepend the path of the playlist
            {
                s = abs_path + "/" + s;
                fi.setFile( s );
            }

			if(fi.isDir())
				(void) new QmDirItem(this, s);
			else
				(void) new QmSongItem(this, s);
		}
    }
    else
        return false;
    return true;
}



/*!
  Increases the global counter.  The counter is used for sorting the songs
  in the playlist.

  \return The last counter value, i.e. before it was increased.
*/

int
QmPlayList::count()
{
    return s_PlayListCounter++;
}


/*!
  This function will traverse the list downwards from the current playing
  song to the \e song first found below.  Directories and bad songs will
  be skipped.
  
  \return The next song, or null if the end is reached.
  \sa prevSong()
*/

QmSongItem*
QmPlayList::nextSong()
{
	QmPlayListItem *nextitem = 0;
	
    if(m_pPlayingItem != 0)
		nextitem = dynamic_cast<QmPlayListItem*>(m_pPlayingItem->itemBelow());
	else
		nextitem = dynamic_cast<QmPlayListItem*>(firstChild());

	nextitem = nextSong(nextitem);
		
	if(nextitem == 0)
		return 0;
	else
	{
		QmSongItem *nextsong = dynamic_cast<QmSongItem*>(nextitem);
		
		if(m_pPlayingItem)
			m_pPlayingItem->setPlaying(false);
		setSelected(nextsong, true);

		return nextsong;
	}
}


/*!
  This function will traverse the list upwards from the current playing
  song to the \e song first found above.  Directories and bad songs will
  be skipped.
  
  \return The previous song, or null if the top is reached.
  \sa nextSong()
*/

QmSongItem*
QmPlayList::prevSong()
{
	QmPlayListItem *previtem = 0;

    if(m_pPlayingItem != 0)
		previtem = dynamic_cast<QmPlayListItem*>(m_pPlayingItem->itemAbove());
	else
		previtem = dynamic_cast<QmPlayListItem*>(firstChild());

	previtem = prevSong(previtem);
	
	if(previtem == 0)
		return 0;
	else
	{
		QmSongItem *prevsong = dynamic_cast<QmSongItem*>(previtem);
		
		if(m_pPlayingItem)
			m_pPlayingItem->setPlaying(false);
		setSelected(previtem, true);

		return prevsong;
	}
}


/*!
  This function is used to search for the next non-bad song starting from
  \a startitem.  Note that if \a startitem is a non-bad song, \e it will be
  the song returned!
  
  \return The song found, or null if none.
  \sa prevSong(QListViewItem*)
*/

QmSongItem*
QmPlayList::nextSong(
	QListViewItem *startitem) const
{
	QmPlayListItem *nextitem = dynamic_cast<QmPlayListItem*>(startitem);

	// Only song items can be played.  If there are folder and sub folder items
	// before the song item, these must be skipped.
	
	while(nextitem != 0)
	{
		if(nextitem->isSong())
        {
            QmSongItem *songitem = dynamic_cast<QmSongItem*>(nextitem);
            if ( ! songitem->isBad() )
				return songitem;
        }
        nextitem = dynamic_cast<QmPlayListItem*>(nextitem->itemBelow());
	}

	return 0;
}


/*!
  This function is used to search for the previous non-bad song starting from
  \a startitem.  Note that if \a startitem is a non-bad song, \e it will be
  the song returned!

  \return The song found, or null if none.
  \sa nextSong(QListViewItem*)
*/

QmSongItem*
QmPlayList::prevSong(
	QListViewItem *startitem) const
{
	QmPlayListItem *previtem = dynamic_cast<QmPlayListItem*>(startitem);
	
	// Only song items can be played.  If there are folder and sub folder items
	// before the song item, these must be skipped.
		
	while(previtem != 0)
	{
		if(previtem->isSong())
        {
            QmSongItem *songitem = dynamic_cast<QmSongItem*>(previtem);
            if ( ! songitem->isBad() )
				return songitem;
        }
        previtem = dynamic_cast<QmPlayListItem*>(previtem->itemAbove());
	}

	return 0;
}


/*!
  Sets the item \a item to be selected if \a selected is true.
  If \a selected is false, the item will be deselected.

  This function also emits enableNext(), disableNext(), enablePrev()
  and disablePrev() signals when appropriate.
 */

void
QmPlayList::setSelected(
	QListViewItem *item,
	bool selected)
{
	if(nextSong(item->itemBelow()) == 0)
		emit disableNext();
	else
		emit enableNext();

	if(prevSong(item->itemAbove()) == 0)
		emit disablePrev();
	else
		emit enablePrev();

	QListView::setSelected(item, selected);
}


/*!
  If \a item is a non-bad song item, give it playing status and remove the playing
  status from the song currently having it, if any.  If \a item is marked as
  a bad song or is not a song, this function will do nothing.
*/

void
QmPlayList::setPlaying(
	QListViewItem *item)
{
	QmSongItem *song = dynamic_cast<QmSongItem*>(item);

	if(song == 0)
		return;

	if(song->isBad())
		return;
	
	if(m_pPlayingItem)
		m_pPlayingItem->setPlaying(false);

	song->setPlaying(true);
	m_pPlayingItem = song;
}


/*!
  This function will return the playing song.  Note that the playing song
  may not actually be playing.  It can be paused or even stopped, but it
  the playing song is the song that will be played if play is pressed.

  \return The playing song, or null if no song has been played yet.
  \todo I don't like this function.  The only reason for its existance is to save the
  name of the playing item when quitting.
*/

QmSongItem*
QmPlayList::playingSong()
{
	return m_pPlayingItem;
}


/*!
  Clears the playlist.  The playing item will be cleared.
*/

void
QmPlayList::clear()
{
	m_pPlayingItem = 0;
	QListView::clear();
}


/*!
  Removes the current item, if any.

  If the current item is the playing item, the next song to play will
  be returned.  If there is no next song to play or there is no need to
  play a new song, null will be returned.

  If there is no next song to play AND the current item is the playing
  item, \a stop_playing will be set to true.  The callee should stop
  playing as the playing item has been deleted, but there was no other
  song to return for playing.  \a stop_playing will be false in all other
  cases.

  \return The new song to play if any, otherwise null.
*/

QmSongItem*
QmPlayList::removeCurrentItem(
	bool *stop_playing)
{
    QmPlayListItem *item = dynamic_cast<QmPlayListItem*>(currentItem());

	*stop_playing = false;
	
	if(item == 0)
		return 0;

	if(item->isSong())
	{
		// The item to remove is a song.

		QmSongItem *song = 0;
		
		if(item == m_pPlayingItem)
		{
			song = nextSong();

			if(song == 0)
			{
				m_pPlayingItem = 0;
				*stop_playing = true;
			}
			else
				setPlaying(song);
		}

		delete item;
		return song;
	}

	// The item to remove is a folder.  We must scan all its
	// children items to check whether the playing item is
	// within the folder, and if so, we must update the pointer,
	// as well as play the next item (below the folder being
	// deleted).
			
	QListViewItem 		*tempitem 	   = item->itemBelow();
	bool 				 within_folder = false;
	
	// Check whether the folder contains the playing item.
	
	while(tempitem != 0)
	{
		if(tempitem->depth() <= item->depth())
			break;
				
		if(tempitem == m_pPlayingItem)
		{
			// Note that we do not break the loop as we want tempitem to point
			// to the last item within the folder so we can continue our
			// search for a new song to play from there.  Otherwise, we would
			// play another song within the folder being deleted!
			
			within_folder = true;
		}

		tempitem = tempitem->itemBelow();				
	}

	if( ! within_folder)
	{
		delete item;
		return 0;
	}
	else
	{
		// The playing item is within the folder being deleted.
		// Find the non-bad playable item below the folder.

		QmSongItem *songitem = nextSong(tempitem);

		if(songitem == 0)
		{
			m_pPlayingItem = 0;
			*stop_playing = true;
		}
		else
		{
			cout << songitem->filePath() << endl;

			setPlaying(songitem);
		}

		delete item;
		return songitem;
	}
}


/*!
  \return True if there is no playing item or the playing item is
  at the beginning.
*/

bool
QmPlayList::atBeginning() const
{
    if ( !m_pPlayingItem )
        return true;
    return prevSong( m_pPlayingItem->itemAbove() ) == 0;
}


/*!
  \return True if there is no playing item or the playing item
  is at the end.
*/

bool
QmPlayList::atEnd() const
{
    if ( !m_pPlayingItem )
        return true;
    return nextSong( m_pPlayingItem->itemBelow() ) == 0;
}
