/* qmsongitem.cpp
 *
 * $Id: qmsongitem.cpp,v 1.15 2000/12/11 02:47:28 mariuss Exp $
 *
 * Apollo sound player: http://www.apolloplayer.org
 * Copyright(C) 2000 Apollo Team.  See CREDITS file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The GNU General Public License is also available online at:
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#include "qmsongitem.h"

#include <qstringlist.h>
#include <qdir.h>
#include <qfileinfo.h>
#include <qpixmap.h>

#include "icons/songitem.xpm"

/*! \file qmsongitem.cpp
  Implementation of QmSongItem class which represents a song in the playlist.
*/

/*
  These static variables are used for maintaining the bad list.  The
  m_BadList variables is the list if file names (including path) that are
  marked as bad.

  The m_ListLoaded will be false for the very first instance created, and
  will result in the list being loaded from disk.  For the next instance
  created, the bad list will already be loaded.

  The m_ItemCount variable is used to count the number of instances there
  are of the class.  When this reaches zero, the bad list is saved to disk.
*/

QStringList 	QmSongItem::m_BadList;
bool 			QmSongItem::m_ListLoaded = false;
int				QmSongItem::m_ItemCount = 0;

/*!
  \class QmSongItem qmsongitem.h
  \brief Represents a song in the play list.

  This class maintains a so-called bad list.  The bad list is a list
  of files which will be skipped during play list playing.  This is
  useful if some files are played incorrectly by the underlaying player,
  even though the files may be fine.
*/


/*!
  Constructs a new song item in the list view \a parent,
  the song being pointed to in \a filename.
*/

QmSongItem::QmSongItem(
    QListView* parent,
    const char* filename,
    int count)
    : QmPlayListItem(parent, count),
    m_File(filename)
{
	setPixmap( 0, QPixmap( songitem_xpm ) );
    m_Playing = false;
	m_ItemCount++;

	if( ! m_ListLoaded)
	{
		m_ListLoaded = true;
		
		QString badlist = QDir::homeDirPath() + "/.apollo/badlist";
		QFile f(badlist);
		if(f.open( IO_ReadOnly ))
		{
			QDataStream str( &f );
			str >> m_BadList;
		}
	}

	if(m_BadList.contains(filename) > 0)
		m_IsBad = true;
	else
		m_IsBad = false;
}


/*!
  Constructs a new song item as a child of the folder \a parent,
  the song being pointed to in \a filename.
*/

QmSongItem::QmSongItem(
    QmPlayListItem* parent,
    const char* filename,
    int count)
    : QmPlayListItem(parent, count),
    m_File(filename)
{
	setPixmap( 0, QPixmap( songitem_xpm ) );
    m_Playing = false;
	m_ItemCount++;

	if( ! m_ListLoaded)
	{
		m_ListLoaded = true;

		QString badlist = QDir::homeDirPath() + "/.apollo/badlist";
		QFile f(badlist);
		if(f.open( IO_ReadOnly ))
		{
			QDataStream str( &f );
			str >> m_BadList;
		}
	}

	if(m_BadList.contains(filename) > 0)
		m_IsBad = true;
	else
		m_IsBad = false;
}


/*!
  Destroys the song item.

  If this is the last song item instance, the bad list will be saved.
*/

QmSongItem::~QmSongItem()
{
	if(--m_ItemCount == 0)
	{
		QString badlist = QDir::homeDirPath() + "/.apollo/badlist";
		QFile f(badlist);
		f.open( IO_WriteOnly );
		QDataStream str( &f );
		str << m_BadList;
	}
}


/*!

 */

void
QmSongItem::setPlaying(
    bool p)
{
    m_Playing = p;
	repaint();
}


/*!

*/

QString
QmSongItem::text(
    int column) const
{
    switch(column)
    {
    case 0:
        if ( m_Title.isEmpty() )
            return m_File.fileName();
        else
            return m_Title;
	break;

    case 1:
        if ( m_Length.isEmpty() )
//            return QString::number(m_File.size());
			return "";
        else
            return m_Length;
	break;

    default:
	return "-";
	break;
    }
}


/*!

*/

QString
QmSongItem::filePath() const
{
    return m_File.filePath();
}



/*!

*/

void
QmSongItem::paintCell(
    QPainter *p,
    const QColorGroup &cg,
    int column,
    int width,
    int alignment)
{
//	QmPlayListItem::paintCell(p, cg, column, width, alignment);
    p->fillRect(0, 0, width, height(), white);

	if(m_IsBad)
		p->setPen(cg.mid());
	else if(m_Playing)
		p->setPen(blue);
	else
		p->setPen(cg.text());

	if(column == 0)
	{
		p->drawPixmap(listView()->itemMargin(), listView()->itemMargin(),
					  *pixmap(0));
	
		p->drawText(listView()->itemMargin() + pixmap(0)->width() + listView()->itemMargin(),
					height() / 2 - p->fontMetrics().height() / 2,
					width + 1,
					height(),
					alignment,
					text(column));
	}
	else
	{
		p->drawText(listView()->itemMargin(),
					height() / 2 - p->fontMetrics().height() / 2,
					width + 1,
					height(),
					alignment,
					text(column));
	}
}


/*!

 */

void
QmSongItem::paintFocus(
    QPainter *p,
    const QColorGroup &cg,
    const QRect &r)
{
    p->setPen(cg.text());
    QRect rect(r.x() + 1, r.y() + 1, r.width() - 2, r.height() - 2);
    p->drawWinFocusRect(rect);
}


/*!
  \return The save string of the item.  This string can be saved to
  disk, and loaded at a later time to create an identical instance
  of this class.
 */

QString
QmSongItem::save() const
{
    return m_File.filePath();
}


/*!
  This function is used to determine whether an item is a song
  or a folder.

  \return True for items of this class.
*/

bool
QmSongItem::isSong() const
{
	return true;
}


/*!
  Sets the bad mark if \a mark is true, clears it if false.
*/

void
QmSongItem::setMark( bool mark )
{
	m_IsBad = mark;

	if(m_IsBad)
    {
        if ( m_BadList.find( filePath() ) == m_BadList.end() ) // Check if the file already is in the list
            m_BadList.append(filePath());
    }
	else
		m_BadList.remove(filePath());
	
	repaint();
}


/*!
  Sets the title of this song, this used instead of the filename if non-empty.
*/

void
QmSongItem::setSongTitle( const QString &title )
{
    m_Title = title;
}

/*!
  Sets the length of the song in minutes and seconds, this used instead of the file size if non-empty.
*/

void
QmSongItem::setSongLength( const QString &len )
{
    m_Length = len;
}

QString
QmSongItem::key( int col, bool ) const
{
    if ( depth() == 0 )
    {
        return sortCounterString();
    }
    else
    {
        if ( col == 0 )
            return QString( "b" ) + text( 0 );
        else
            return text( col );
    }
}

bool
QmSongItem::isBad() const
{
    return m_IsBad;
}
