/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of MixMagic.

    Copyright (C) 2000 Jonas Borgstrm <jonas_b@bitsmart.com>.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/mman.h>

#include "mixmagic.h"
#include "index.h"
#include "sample.h"

/*
 * Create a new index.
 */
Index *
index_new (Sample *sample, GnomeAppBar *appbar)
{
	Index *index;
	struct stat stat1, stat2;

	index = g_new0 (Index, 1);

	index->filename = g_strdup_printf ("%s%c.mixmagic%c%s.idx", g_get_home_dir (), 
					   G_DIR_SEPARATOR, 
					   G_DIR_SEPARATOR, 
					   g_basename (sample->filename));

	fstat (sample->fd, &stat2);

	if (stat (index->filename, &stat1) == -1 ||
	    stat2.st_mtime > stat1.st_mtime) {

		index_create (index, sample, appbar);
	} else {

		if ((index->fd = open (index->filename, O_RDONLY)) == -1)
			g_warning ("Can't open index file '%s'\n", index->filename);
	}

	/* Get filesize */
	index->map_size = lseek (index->fd, 0, SEEK_END);
	
	index->data = mmap (0, index->map_size, 
			    PROT_READ, MAP_SHARED, 
			    index->fd, 0);

	g_assert (index->data != (gpointer)-1);

	return index;
}

gboolean
index_create (Index *index, Sample *sample, GnomeAppBar *appbar)
{
	FILE *out;
	gint8 data;
	gint16 *input;
	gchar *dirname, *msg;
	guint8 count = 0;

	/* Try to create the index directory */
	dirname = g_strdup_printf ("%s%c.mixmagic", g_get_home_dir (), G_DIR_SEPARATOR);
	mkdir (dirname, S_IRWXU);
	g_free (dirname);

	if ((out = fopen (index->filename, "w")) == NULL) {

		g_error ("Failed to create '%s'\n", index->filename);
	}
	input = (gint16 *)sample->data;
	
	msg = g_strdup_printf ("Creating index file: %s", index->filename);
	gnome_appbar_push (appbar, msg);
	g_free (msg);

	while ((gpointer)input < (gpointer)(sample->data + sample->length)) {

		if (count++ == 0) {

			gnome_appbar_set_progress (appbar, 
						   (float)((int)input - (int)sample->data) / (float)sample->length);
			gtk_main_iteration_do (FALSE);
		}
		data = *input / 0xff;
		fwrite (&data, sizeof (gint8), 1, out);
		data = *(input+1) / 0xff;
		fwrite (&data, sizeof (gint8), 1, out);
		input += 16 * 2;
	}
	fclose (out);

	/* Reset the appbar */
	gnome_appbar_set_progress (appbar, 0);
	gnome_appbar_pop (appbar);

	/* Reopen the newly created index */

	if ((index->fd = open (index->filename, O_RDONLY)) == -1) {

		g_error ("Faild to open index '%s', exiting\n", index->filename);
	}

	return TRUE;

}

void 
index_destroy (Index *index)
{
	if (index->filename)
		g_free (index->filename);

	if (munmap (index->data, index->map_size) == -1)
		g_warning ("munmap (%p, %d) failed\n", 
			   index->data, 
			   index->map_size);
	if (index->fd)
		close (index->fd);
	
	g_free (index);
}

guint8
index_get_average (Index *index, gint offset, gint start, gint length, gint loop_length)
{
	guint value = 0, i;
	gint8 *data;

	data = index->data + offset;

	for (i=0; i <length;i++) {
		while ((guint)data >= (guint)index->data + start + loop_length)
			data -= loop_length;
		
		value += ABS (*data) + ABS(*(data+1));
		data += 2;
	}
	return value / length;
}
